@@ -18,6 +18,7 @@
  */
 
 #include "libinit.h"
+#include "statx.h"
 
 #define USAGE		"Usage:\n"\
 			"\t%s [-v] [-k] [-p pid_file] /full/path/to/program\n" \
@@ -31,10 +32,11 @@ int main(int argc, char **argv)
     int blog = 0;
 #endif
     struct stat st;
-    PROC * list;
+    list_t *list;
     char *fullname = NULL, * basename = NULL;
     char *pid_file = NULL, *ignore_file = NULL;
-    char *root = NULL, *iargv[argc];
+    char *root = NULL;
+    extension char *iargv[argc];
     char *posixa, *posixb;	/* Don't fool me with posix correct */
     int quiet = 1, iargc = 0;
     unsigned short flags = (DAEMON|PIDOF|NZOMBIE);
@@ -59,7 +61,7 @@ int main(int argc, char **argv)
     posixa = getenv("_POSIX_OPTION_ORDER"); unsetenv("_POSIX_OPTION_ORDER");
     posixb = getenv("POSIXLY_CORRECT");     unsetenv("POSIXLY_CORRECT");
     opterr = 0;
-    while ((c = getopt(argc, argv, "c:kp:nNhqvzLi:")) != -1) {
+    while ((c = getopt(argc, argv, "c:kp:nNhqvzLi:x")) != -1) {
 	switch (c) {
 	    case 'c':
 		if (optarg && optarg[0] != '-' && !root) {
@@ -85,6 +87,9 @@ int main(int argc, char **argv)
 	    case 'N':
 		init_nfs();
 		break;
+	    case 'x':
+		flags |= STSCRPT;
+		break;
 	    case 'z':
 		flags &= ~NZOMBIE;
 		break;
@@ -92,7 +97,6 @@ int main(int argc, char **argv)
 		/* Allocate here: address optarg (current *argv) isn't freeable */
 		if (optarg && !pid_file) {
 		    pid_file = xstrdup(optarg);
-		    pid_forced = true;
 		} else
 		    error(WRGSYNTAX, "Option -p requires pid file to read pid from\n");
 		break;
@@ -137,7 +141,8 @@ int main(int argc, char **argv)
     if (*fullname != '/')
 	flags |= KSHORT;
     else if (check4nfs(fullname))
-	flags |= (KSHORT|KBASE);;
+	flags |= (KSHORT|KBASE);
+    clear_nfs(); 
 
     if (!pid_file) {		/* the default pid file */
 	if (root) {
@@ -148,11 +153,12 @@ int main(int argc, char **argv)
 	    pid_file = (char*) xmalloc(DEFPIDLEN+strlen(basename)+1);
 	    pid_file = strcat(strcat(strcpy(pid_file,DEFPIDDIR),basename),DEFPIDEXT);
 	}
-    }
+    } else
+	pid_forced = true;
 
     /* Check and verify the pid file */
     errno = 0;
-    if (stat(pid_file, &st) < 0) {
+    if (statn(pid_file, STATX_SIZE, &st) < 0) {
 	if (errno != ENOENT) {
 	    /* An other error like permission or HW problem */
 	    warn("Can not stat %s: %s\n", pid_file, strerror(errno));
@@ -163,28 +169,33 @@ int main(int argc, char **argv)
 	     * the proc is dead if the specified pid can not be veryfied.
 	     */
 	    if (remember_pids(pid_file,fullname,root,flags) < 0)
-		exit(LSB_PROOFX);
+		exit(LSB_STATUS_PROOFX);
 
-	    if (!remember)
+	    if (list_empty(&remember))
 		exit(LSB_STATUS_NOPROC);	/* New LSB: no pid file is no job */
 	}
-	/* No pid file means that we have to search in /proc/ */
 	free(pid_file);
 	pid_file = NULL;
+
+	/* No pid file means that we have to search in /proc/ */
     }
 
     if (pid_file && !st.st_size) {
 	warn("Empty pid file %s for %s\n", pid_file, fullname);
 
-	/* No pid file means that we have to search in /proc/ */
 	free(pid_file);
 	pid_file = NULL;
+
+	if (pid_forced)
+	    exit(LSB_STATUS_NOPROC);
+
+	/* No pid file means that we have to search in /proc/ */
     }
 
     /* Check and verify the ignore file */
     for (c = 0; (c < iargc) && (ignore_file = iargv[c]); c++) {
 	errno = 0;
-	if (stat(ignore_file, &st) < 0) {
+	if (statn(ignore_file, STATX_SIZE, &st) < 0) {
 	    if (errno != ENOENT)
 		warn("Can not stat %s: %s\n", ignore_file, strerror(errno));
 	    continue;
@@ -203,8 +214,12 @@ int main(int argc, char **argv)
 	if (pid_file) {		  /* The case of having a pid file */
 	    if (verify_pidfile(pid_file,fullname,root,flags,false) < 0)
 		exit(LSB_STATUS_PROOFX);
+	    if (iargc)
+		clear_pids();
 	}
-	if (!remember) {	  /* No process found with pid file */
+	if (list_empty(&remember)) {	  /* No process found with pid file */
+	    if (pid_forced)
+		exit(LSB_STATUS_PROOFX);
 	    if (pidof(fullname,root,flags) < 0)
 		exit(LSB_STATUS_PROOFX);
 	}
@@ -220,16 +235,17 @@ int main(int argc, char **argv)
     clear_pids();		/* Remove all pids which should be ignored */
 
     num = 0;	/* If quiet we could test 'remember' and exit appropiate */
-    for(list = remember; list; list = list->next) {
+    list_for_each(list, &remember) {
+	PROC *proc = list_entry(list, PROC);
 	if (!quiet) {
 	    if (num) putchar(' ');
-	    printf("%d", (int)list->pid);
+	    printf("%ld", (long int)proc->pid);
 	}
 #ifdef USE_BLOGD
 	if (bootlog(B_NOTICE, "%s: %s ", we_are, fullname) == 0) {
 	    blog++;
 	    if (num) bootlog(-1, " ");
-	    bootlog(-1, "%d", (int)list->pid);
+	    bootlog(-1, "%ld", (long int)proc->pid);
 	}
 #endif
 	num++;
