@@ -22,6 +22,7 @@
  */
 
 #include "libinit.h"
+#include "statx.h"
 
 #define DEFSIG		"TERM"
 #define OTHERSIG	"HUP"
@@ -38,17 +39,18 @@ static int quiet = 1, num;
 
 int main(int argc, char **argv)
 {
-    extern char * we_are;
     int c, snum;
     struct stat st;
-    PROC * list;
+    list_t *list;
     char *fullname = NULL, *basename = NULL;
     char *pid_file = NULL, *ignore_file = NULL;
-    char *root = NULL, *iargv[argc];
+    char *root = NULL;
+    extension char *iargv[argc];
     char *posixa, *posixb;	/* Don't fool me with posix correct */
     int process_group = 0, group_leader = 0, wait = 5, iargc = 0;
     unsigned short flags = (KILL|PIDOF|KSTOP);
     boolean pid_forced = false;
+    boolean sig_forced = true;
 
     we_are = base_name(argv[0]);
     openlog (we_are, LOG_OPTIONS, LOG_FACILITY);
@@ -77,14 +79,16 @@ int main(int argc, char **argv)
 	    int tmp, len = strlen(sig);
 	    sig++;
 	    if ( (tmp = atoi(sig)) > 0 && tmp < NSIG ) {
-		bzero(sig, len);
+		memset(sig, 0, len);
 		*sig = 'q';		/* set dummy option -q */
 		snum = tmp;
+		sig_forced = false;
 		break;
 	    } else if ( (tmp = signame_to_signum(sig)) > 0 ) {
-		bzero(sig, len);
+		memset(sig, 0, len);
 		*sig = 'q';		/* set dummy option -q */
 		snum = tmp;
+		sig_forced = false;
 		break;
 	    }
 	}
@@ -93,7 +97,7 @@ int main(int argc, char **argv)
     posixa = getenv("_POSIX_OPTION_ORDER"); unsetenv("_POSIX_OPTION_ORDER");
     posixb = getenv("POSIXLY_CORRECT");     unsetenv("POSIXLY_CORRECT");
     opterr = 0;
-    while ((c = getopt(argc, argv, "c:p:gGnNhlvqt:Li:")) != -1) {
+    while ((c = getopt(argc, argv, "c:p:gGnNhlvqt:Li:x")) != -1) {
 	switch (c) {
 	    case 'c':
 		if (optarg && optarg[0] != '-' && !root) {
@@ -132,11 +136,13 @@ int main(int argc, char **argv)
 	    case 'N':
 		init_nfs();
 		break;
+	    case 'x':
+		flags |= STSCRPT;
+		break;
 	    case 'p':		/* Changed from -f to -p to fit startproc and LSB */
 		/* Allocate here: address optarg (current *argv) isn't freeable */
 		if (optarg && !pid_file) {
 		    pid_file = xstrdup(optarg);
-		    pid_forced = true;
 		} else
 		    error(LSB_WRGSYN,"Option -p requires pid file to read pid from\n");
 		break;
@@ -156,7 +162,6 @@ int main(int argc, char **argv)
 	    case 'h':
 		error(0, USAGE);
 		break;
-		break;
 	    default:
 		break;
 	}
@@ -187,6 +192,7 @@ int main(int argc, char **argv)
 	flags |= KSHORT;
     else if (check4nfs(fullname))
 	flags |= (KSHORT|KBASE);
+    clear_nfs();
 
     if (!pid_file) {            /* the default pid file */
 	if (root) {
@@ -197,11 +203,12 @@ int main(int argc, char **argv)
 	    pid_file = (char*) xmalloc(DEFPIDLEN+strlen(basename)+1);
 	    pid_file = strcat(strcat(strcpy(pid_file,DEFPIDDIR),basename),DEFPIDEXT);
 	}
-    }
+    } else
+	pid_forced = true;
 
     /* Check and verify the pid file */
     errno = 0;
-    if (stat(pid_file, &st) < 0) {
+    if (statn(pid_file, STATX_SIZE, &st) < 0) {
 	if (errno != ENOENT) {
 	    /* An other error like permission or HW problem */
 	    warn("Can not stat %s: %s\n", pid_file, strerror(errno));
@@ -214,20 +221,25 @@ int main(int argc, char **argv)
 	    if (remember_pids(pid_file,fullname,root,flags) < 0)
 		exit(LSB_PROOFX);
 
-	    if (!remember)
+	    if (list_empty(&remember))
 	        exit(LSB_NOPROC);	/* New LSB: no pid file is no job */
 	}
-	/* No pid file means that we have to search in /proc/ */
 	free(pid_file);
 	pid_file = NULL;
+
+	/* No pid file means that we have to search in /proc/ */
     }
 
     if (pid_file && !st.st_size) {
 	warn("Empty pid file %s for %s\n", pid_file, fullname);
 
-	/* No pid file means that we have to search in /proc/ */
 	free(pid_file);
 	pid_file = NULL;
+
+	if (pid_forced)
+	    exit(LSB_STATUS_NOPROC);
+
+	/* No pid file means that we have to search in /proc/ */
     }
 
     if (pid_file) {		/* The case of having a pid file */
@@ -241,7 +253,7 @@ int main(int argc, char **argv)
     /* Check and verify the ignore file */
     for (c = 0; (c < iargc) && (ignore_file = iargv[c]); c++) {
 	errno = 0;
-	if (stat(ignore_file, &st) < 0) {
+	if (statn(ignore_file, STATX_SIZE, &st) < 0) {
 	    if (errno != ENOENT)
 		warn("Can not stat %s: %s\n", ignore_file, strerror(errno));
 	    continue;
@@ -259,7 +271,7 @@ int main(int argc, char **argv)
 	clear_pids();		/* Remove all pids which should be ignored */
 
     /* Do main work */
-    if (!remember) {
+    if (list_empty(&remember)) {
 	/* killing a none existing process is already success.
 	 * Nevertheless LSB says that we should note that if
 	 * a signal is explicit given.
@@ -268,8 +280,10 @@ int main(int argc, char **argv)
     }
 
     num = 0;
-    for(list = remember; list; list = list->next)
-	do_kill(basename, list->pid, snum, group_leader, process_group);
+    list_for_each(list, &remember) {
+	PROC *proc = list_entry(list, PROC);
+	do_kill(basename, proc->pid, snum, group_leader, process_group);
+    }
 
     if (snum == SIGTERM || snum == SIGKILL) {
 	int partsec = 5*wait;	/* We look 5 times within a second */
@@ -283,7 +297,7 @@ again:
 	if (check_pids(fullname,root,flags) < 0)
 	    exit(LSB_PROOFX);
 
-	if (!remember)		/* success */
+	if (list_empty(&remember))		/* success */
 	    goto success;
 
 	fflush(stdout);
@@ -293,14 +307,19 @@ again:
 	    goto again;
 	}
 
-	if (snum == SIGKILL)
+	if (snum == SIGKILL)	/* SIGKILL was specified on the command line */
+	    goto badterm;
+
+	if (!sig_forced)	/* SIGTERM was specified on the command line */
 	    goto badterm;
 
 	if (check_pids(fullname,root,flags) < 0)
 	    exit(LSB_PROOFX);
 
-	for(list = remember; list; list = list->next)
-	    do_kill(basename, list->pid, SIGKILL, group_leader, process_group);
+	list_for_each(list, &remember) {
+	    PROC *proc = list_entry(list, PROC);
+	    do_kill(basename, proc->pid, SIGKILL, group_leader, process_group);
+	}
 
 	/* Do we have killed them? */
 
@@ -308,7 +327,7 @@ again:
 	if (check_pids(fullname,root,flags) < 0)
 	    exit(LSB_PROOFX);
 
-	if (remember)
+	if (!list_empty(&remember))
 	    goto badterm;
 
 success:
@@ -343,7 +362,7 @@ static int do_kill(const char *inname, const pid_t proc, const int sig,
 	if ((target = -getpgid(proc)) >= 0) {
 	    if (errno != ESRCH)
 		warn("Can not signal %s to process with pid %d: %s\n",
-			sys_signame[sig], (int)proc, strerror(errno));
+			signum_to_signame(sig), (int)proc, strerror(errno));
 	    exit(LSB_PROOFX);
 	}
     } else if (process_group)
@@ -359,7 +378,7 @@ static int do_kill(const char *inname, const pid_t proc, const int sig,
     if (kill(target, sig) < 0) {
 	if (errno != ESRCH) {
 	    warn("Can not signal %s to process with pid %d: %s\n",
-		    sys_signame[sig], (int)proc,  strerror(errno));
+		    signum_to_signame(sig), (int)proc,  strerror(errno));
 	    exit(LSB_FAILED);
 	}
     }
@@ -368,7 +387,7 @@ static int do_kill(const char *inname, const pid_t proc, const int sig,
 			   the environment of the current processes */
     if (!quiet) {
 	if (num++) putchar(' ');
-	printf("SIG%s %s(%d)",sys_signame[sig],inname,(int)proc);
+	printf("SIG%s %s(%d)",signum_to_signame(sig),inname,(int)proc);
     }
 #endif
     return 0;
