@@ -19,6 +19,11 @@
  * 2000/11/10 Werner Fink: LSB specs, logging
  */
 
+#ifndef _LIBINIT_H
+#define _LIBINIT_H
+#ifndef _GNU_SOURCE
+# define _GNU_SOURCE
+#endif
 #include <stdlib.h>
 #include <stdio.h>
 #include <dirent.h>
@@ -29,6 +34,7 @@
 #include <utmp.h>
 #include <pwd.h>
 #include <mntent.h>
+#include <libgen.h>
 #include <string.h>
 #include <unistd.h>
 #include <fcntl.h>
@@ -40,6 +46,10 @@
 #ifdef USE_BLOGD
 # include <libblogger.h>
 #endif
+#include "lists.h"
+#ifndef O_CLOEXEC
+# define O_CLOEXEC 0	
+#endif
 
 /*
  * LSB specs:
@@ -81,8 +91,8 @@
 #define WRGSYNTAX		102	/* usage etc. */
 #define NOPIDREAD		101	/* trouble */
 
-#define LSB_STATUS_PROOF	((errno == EPERM) ? LSB_NOPERM : NOPIDREAD )
-#define LSB_STATUS_PROOFX	((errno == ENOENT) ? LSB_NOENTR : LSB_STATUS_PROOF )
+#define LSB_STATUS_PROOF	((errno == EPERM || errno == EACCES) ? LSB_NOPERM : LSB_STATUS_ISDEAD )
+#define LSB_STATUS_PROOFX	((errno == ENOENT) ? (flags & KILL) ? LSB_NOENTR : 4 : LSB_STATUS_PROOF )
 
 #define LOG_OPTIONS	(LOG_ODELAY|LOG_CONS)
 #define LOG_FACILITY	LOG_LOCAL7
@@ -96,6 +106,7 @@
 #define FLWLINK 0x0040
 #define KSTOP   0x0080
 #define KBASE   0x0100
+#define STSCRPT 0x0200
 
 #define  MAXENV  20
 #define  CMDLLEN MAXNAMLEN	/* The string length of /proc/12345/cmdline\0\0 + 1 */
@@ -106,8 +117,7 @@
 #endif
 #define  DEFPIDEXT ".pid"
 #define  DEFPIDLEN 14		/* The string length of /var/run/.pid + 1 */
-
-typedef enum _boolean {false, true} boolean;
+#define  COMM_LEN 15		/* The lenght of the task command name in /proc/<pid>/stat */
 
 extern char **environ;
 extern char     * newenvp[];
@@ -128,29 +138,28 @@ extern void clear_pids (void);
 extern void error(int stat, const char *fmt, ...);
 extern void warn(const char *fmt, ...);
 extern int rlstat(char ** file, struct stat *st, const unsigned short flag);
+extern char* expandpath(const char * path);
 extern void init_nfs(void);
-boolean check4nfs(const char * path);
+extern void clear_nfs(void);
+extern void getproc(void);
+extern boolean check4nfs(const char * path);
+extern size_t dirdepth(const char *const path);
 
 /* Feature */
-extern char *sys_signame [];
-extern char *sys_sigalias[];
-extern void init_signames();
-
 extern int signame_to_signum (const char *sig);
+extern const char * signum_to_signame (const int sig);
+extern void list_signames(void);
 
+#if 0
 /* Used in killproc.c only once to list the signal names */
 static inline void list_signames(void)
 {
     int n, l;
 
-    init_signames();
     for (n = 1, l = 0; n < NSIG+1; n++) {
-	if (sys_signame [n]) {
-	    printf("%2d) SIG%-9s", n, sys_signame [n]);
-	    l++;
-	}
-	if (sys_sigalias[n]) {
-	    printf("%2d) SIG%-9s", n, sys_sigalias[n]);
+	const char * signame = signum_to_signame(n);
+	if (signame) {
+	    printf("%2d) SIG%-9s", n, signame);
 	    l++;
 	}
 	if (!(l % 4))
@@ -159,17 +168,17 @@ static inline void list_signames(void)
     if (l % 4)
 	putc('\n', stdout);
 }
+#endif
 
 typedef struct _proc_
 {
-    struct _proc_ *next;	/* Pointer to next struct. */
-    struct _proc_ *prev;	/* Pointer to previous st. */
-    pid_t pid;			/* Process ID.             */
-    pid_t sid;			/* Session ID.             */
+    list_t this;
+    pid_t   pid;			/* Process ID. */
+    pid_t   sid;			/* Session ID. */
 } PROC;
 
-extern PROC * remember;
-extern PROC * doignore;
+extern list_t remember;
+extern list_t doignore;
 
 /* Inlined functions: just like macros */
 
@@ -209,40 +218,12 @@ static inline char * base_name ( const char * full )
 
 static inline char * swap_name ( const char * base )
 {
-    size_t len = strlen(base) + 2 + 1;
-    char *swap = (char*)xmalloc(len);
-
-    return strcat(strcat(strcpy(swap,"("),base),")");
-}
-
-static inline char * exe_entry ( char *buf, const char *d_name )
-{
-    return strcat(strcat(strcpy(buf,"/proc/"),d_name),"/exe");
-}
-
-static inline char * stat_entry ( char *buf, const char *d_name )
-{
-    return strcat(strcat(strcpy(buf,"/proc/"),d_name),"/stat");
-}
-
-static inline char * statm_entry ( char *buf, const char *d_name )
-{
-    return strcat(strcat(strcpy(buf,"/proc/"),d_name),"/statm");
-}
-
-static inline char * env_entry ( char *buf, const char *d_name )
-{
-    return strcat(strcat(strcpy(buf,"/proc/"),d_name),"/environ");
-}
-
-static inline char * cmd_entry ( char *buf, const char *d_name )
-{
-    return strcat(strcat(strcpy(buf,"/proc/"),d_name),"/cmdline");
-}
-
-static inline char * pid_entry ( char *buf, const char *pid )
-{
-    return strcat(strcpy(buf,"/proc/"),pid);
+    size_t len = strlen(base);
+    char *swap;
+    if (len > COMM_LEN)
+	len = COMM_LEN;
+    swap = (char*)xmalloc(len + 2 + 1);
+    return strcat(strncat(strcpy(swap,"("),base, COMM_LEN),")");
 }
 
 extern void addnewenv ( const char * name, const char * entry );
@@ -275,6 +256,12 @@ static inline void set_newenv(const char * fullname)
     else
 	addnewenv("SHELL","/bin/sh");
 
+    if ( (tmp = getenv("LISTEN_PID")) != (char*)0 )
+	addnewenv("LISTEN_PID",tmp);
+
+    if ( (tmp = getenv("LISTEN_FDS")) != (char*)0 )
+	addnewenv("LISTEN_FDS",tmp);
+
     if ( (tmp = getenv("RUNLEVEL")) != (char*)0 )
 	addnewenv("RUNLEVEL",tmp);
 
@@ -322,31 +309,21 @@ static inline void do_list(const pid_t pid, const pid_t sid, const boolean ignor
     p->pid = pid;
     p->sid = sid;
     if (ignore) {
-        if (doignore)
-	    doignore->prev = p;
-        p->next = doignore;
-        p->prev = (PROC*)0;
-        doignore = p;
+	append(p, doignore);
     } else {
-        if (remember)
-	    remember->prev = p;
-        p->next = remember;
-        p->prev = (PROC*)0;
-        remember = p;
+	append(p, remember);
     }
-
-    return;
 }
 
 static inline boolean check_ignore(const pid_t pid)
 {
-    PROC * q, * m = doignore;
+    list_t *m;
 
     if (pid <= 1)
 	goto out;
 
-    for (q = doignore; m; q = m) {
-	m = q->next;
+    list_for_each(m, &doignore) {
+	PROC *q = list_entry(m, PROC);
         if (pid == q->pid)
 	    return true;
     }
@@ -356,28 +333,20 @@ out:
 
 static inline void clear_ignore(void)
 {
-    PROC * p, *l, * n = remember;
+    list_t *n, *l;
 
-    if (!doignore)
+    if (list_empty(&doignore))
 	return;
 
-    for (p = remember; n; p = n) {
-	l = p->prev;
-	n = p->next;
+    list_for_each_safe(n, l, &remember) {
+	PROC *p = list_entry(n, PROC);
 
 	if (!check_ignore(p->pid))
 	    continue;
 
 	/* Remove this entry in remember because we ignore it */
-	if (p == remember) {
-	    if (n) n->prev = (PROC*)0;
-	    remember = n;
-	    free(p);
-	} else if (l) {
-	    if (n) n->prev = l;
-	    l->next = n;
-	    free(p);
-	}
+	delete(n);
+	free(p);
     }
 }
-/* libinit.h ends here */
+#endif /* _LIBINIT_H */
