@@ -21,16 +21,23 @@
  *                         do not fork in case of start_daemon
  */
 
-#include "libinit.h"
 #include <sys/time.h>
 #include <sys/resource.h>
+#include <sys/inotify.h>
 #include <sys/ioctl.h>
+#include <sys/syscall.h>
 #include <argz.h>
 #include <grp.h>
+#include "libinit.h"
+#include "statx.h"
+
+#ifndef  SD_LISTEN_FDS_START
+# define SD_LISTEN_FDS_START	3
+#endif
 
 #define USAGE		"Usage:\n"\
 			"    %s [-f] [-L] [[-n ]+/-<prio>] [-s] [-t sec|-T sec] [-u uid] [-g gid] [-v] [-e] \\\n"\
-			"        [-l log|-q|-d] [-p pid_file] [-i ignore_file] [-c root] /path/to/executable [args]\n"
+			"        [-l log|-q|-d] [-p pid_file] [-i ignore_file] [-c root] [-w|-W list] /path/to/executable [args]\n"
 #define USAGE_SD	"Usage:\n"\
 			"    %s [-f] [-L] [-n +/-<prio>] [-u uid] [-g gid] [-v] [-e] \\\n"\
 			"        [-l log|-q|-d] [-p pid_file] [-i ignore_file] [-c root] /path/to/executable [args]\n"
@@ -38,19 +45,25 @@
 static int do_start(const char *name, char *argv[], const char* log_file,
 		   const int nicelvl, const int env, const char* root, unsigned short flags);
 static void closefds(FILE *not);
+static void waiton(const char *list);
+static int get_sd_listen_fds(void);
+static void fwd_sd_listen_pid(void);
 
-static int quiet = 1, supprmsg = 0, sess = 0, seconds = 0, sigchld = 0, force = 0, dialog = 0;
+static int quiet = true, supprmsg = false, sess = false, seconds = false;
+static int sigchld = false, force = false, dialog = false;
 static struct passwd *user = NULL;
-struct group *grp = NULL;
+static struct group *grp = NULL;
 static int syslogd = 0;
 static int sdaemon = 0;
+static int wpopts = WNOHANG|WUNTRACED;
+static char *wlist = NULL;
 
 static volatile sig_atomic_t signaled = 0;
-static void (*save_sigquit) = SIG_DFL;
+static sighandler_t save_sigquit = SIG_DFL;
 static void sig_quit(int nsig)
 {
     (void)signal(nsig, save_sigquit);
-    signaled = 1;
+    signaled = true;
 }
 
 static void sig_chld(int nsig)
@@ -61,13 +74,46 @@ static void sig_chld(int nsig)
     seconds = 0;
 }
 
+enum {
+    IOPRIO_CLASS_NONE,
+    IOPRIO_CLASS_RT,
+    IOPRIO_CLASS_BE,
+    IOPRIO_CLASS_IDLE,
+};
+
+enum {
+    IOPRIO_WHO_PROCESS = 1,
+    IOPRIO_WHO_PGRP,
+    IOPRIO_WHO_USER,
+};
+
+#define IOPRIO_CLASS_SHIFT	(13)
+#define IOPRIO_PRIO_MASK	((1UL << IOPRIO_CLASS_SHIFT) - 1)
+
+#define IOPRIO_PRIO_CLASS(mask)	((mask) >> IOPRIO_CLASS_SHIFT)
+#define IOPRIO_PRIO_DATA(mask)	((mask) & IOPRIO_PRIO_MASK)
+#define IOPRIO_PRIO_VALUE(class, data)	(((class) << IOPRIO_CLASS_SHIFT) | data)
+
+static inline int ioprio_set(int which, int who, int ioprio)
+{
+    return syscall(SYS_ioprio_set, which, who, ioprio);
+}
+
+static int schedclass = -1, scheddata = 4;
+static void inline ioprio_setpid(pid_t pid, int ioclass, int data)
+{
+    int rc = ioprio_set(IOPRIO_WHO_PROCESS, pid, IOPRIO_PRIO_VALUE(ioclass, data));
+    if (rc < 0)
+	error(LSB_PROOF, "ioprio_set failed: %m\n");
+}
+
 int main(int argc, char **argv)
 {
-    extern char * we_are;
     struct stat st;
     char *fullname = NULL, *basename = NULL;
     char *log_file = NULL, *pid_file = NULL, *ignore_file = NULL;
-    char *root = NULL, *iargv[argc];
+    char *root = NULL;
+    extension char *iargv[argc];
     int c, nicelvl = 0, env = 0, iargc = 0;
     unsigned short flags = (DAEMON|PIDOF);
 
@@ -98,7 +144,7 @@ int main(int argc, char **argv)
     }
 
     opterr = 0;
-    while ((c = getopt(argc, argv, "+c:edp:l:hqvsu:g:t:n:fLi:T:")) != -1) { /* `+' is POSIX correct */
+    while ((c = getopt(argc, argv, "+c:edp:l:hqvsu:g:t:n:o:fLi:T:wW:x")) != -1) { /* `+' is POSIX correct */
 	switch (c) {
 	    case 'v':
 		quiet = 0;
@@ -110,13 +156,18 @@ int main(int argc, char **argv)
 		    error(LSB_WRGSYN,"Option -c requires special root directory\n");
 		break;
 	    case 'e':
-		env = 1;
+		env = true;
 		break;
 	    case 'd':
-		dialog = 1;
+		dialog = true;
 		seconds = 15;
 		break;
+	    case 'x':
+		flags |= STSCRPT;
+		break;
 	    case 'p':		/* Former option -f */
+		if (force)
+		    warn("option -p does not work in force mode\n");
 		/* Allocate here: address optarg (current *argv) isn't freeable */
 		if (optarg && optarg[0] != '-' && !pid_file) {
 		    pid_file = xstrdup(optarg);
@@ -124,7 +175,7 @@ int main(int argc, char **argv)
 		    error(LSB_WRGSYN,"Option -p requires pid file to read pid from\n");
 		break;
 	    case 'f':		/* Newer option -f for force start (LSB specs!) */
-		force++;
+		force = true;
 		break;
 	    case 'l':
 		if (optarg && optarg[0] != '-' && !log_file) {
@@ -152,12 +203,58 @@ int main(int argc, char **argv)
 		} else
 		    error(LSB_WRGSYN,"Option -n requires nice level\n");
 		break;
+	    case 'o':
+		if (optarg && optarg[0] != '/' && optarg[0] != '-') {
+		    char *endptr, *class = (char*)0, *level = (char*)0;
+		    int lvl;
+
+		    if ((endptr = strchr(optarg, ','))) {
+			*endptr++ = '\0';
+			level = endptr;
+			class = &optarg[0];
+			if (*level == 'n')
+			    level++;
+			if (*class == 'c')
+			    class++;
+		    } else {
+			if (optarg[0] == 'c')
+			    class = &optarg[1];
+			if (optarg[0] == 'n')
+			    level = &optarg[1];
+			if (optarg[0] >= '0' && optarg[0] <= '3') {
+			    class = &optarg[0];
+			    printf("%s\n", optarg);
+			}
+		    }
+
+		    if (!class && !level)
+			error(LSB_WRGSYN,"Option -o requires a scheduling class and/or scheduling class data level\n");
+
+		    if (class && *class) {
+			lvl = strtol(class, &endptr, 10);
+			if (*endptr != '\0')
+			    error(LSB_WRGSYN,"Class identifier `c' requires a number as scheduling class\n");
+			if (lvl < 0 || lvl > 3)
+			    error(LSB_WRGSYN,"Class identifier `c' is out of range\n");
+			schedclass = lvl;
+		    }
+		    if (level && *level) {
+			lvl = strtol(level, &endptr, 10);
+			if (*endptr != '\0')
+			    error(LSB_WRGSYN,"Class data identifier `n' requires a number as scheduling class data\n");
+			if (lvl < 0 || lvl > 7)
+			    error(LSB_WRGSYN,"Class data identifier `n' is out of range\n");
+			scheddata = lvl;
+		    }
+		} else
+		     error(LSB_WRGSYN,"Option -o requires I/O nice level\n");
+		break;
 	    case 'q':
-	        supprmsg = 1;
+	        supprmsg = true;
 	        break;
 	    case 's':
 		if (sdaemon) goto fail;
-		sess = 1;
+		sess = true;
 		break;
 	    case 'u':
 		if (optarg && optarg[0] != '/' && optarg[0] != '-') {
@@ -220,12 +317,25 @@ int main(int argc, char **argv)
 		    error(LSB_WRGSYN,"Option -t requires number of seconds\n");
 		break;
 	    case 'i':
+		if (force)
+		    warn("option -i does not work in force mode\n");
 		/* Remember: address optarg (current *argv) */
 		if (optarg && optarg[0] != '-') {
 		    iargv[iargc++] = optarg;
 		} else
 		    error(LSB_WRGSYN,"Option -i requires pid file to read pid from\n");
 		break;
+	    case 'w':
+		if (sdaemon) goto fail;
+		wpopts = WUNTRACED;
+		break;
+	    case 'W':
+		if (sdaemon) goto fail;
+		if (optarg && optarg[0] != '-') {
+		    wlist = optarg;
+		} else
+		    error(LSB_WRGSYN,"Option -W requires a file name or list of files separated by colons\n");
+		break;
 	    case '?':
 	    fail:
 		error(LSB_WRGSYN, (sdaemon ? USAGE_SD : USAGE), we_are);
@@ -270,25 +380,34 @@ int main(int argc, char **argv)
 	    pid_file = (char*) xmalloc(DEFPIDLEN+strlen(basename)+1);
 	    pid_file = strcat(strcat(strcpy(pid_file,DEFPIDDIR),basename),DEFPIDEXT);
 	}
-    }
+    } else
+	force = true;
 
     /* Check and verify the pid file */
     errno = 0;
-    if (stat(pid_file, &st) < 0) {
+    if (statn(pid_file, STATX_SIZE, &st) < 0) {
 	if (errno != ENOENT)
 	    warn("Can not stat %s: %s\n", pid_file, strerror(errno));
 
-	/* No pid file means that we have to search in /proc/ */
 	free(pid_file);
 	pid_file = NULL;
+
+	if (force && errno == ENOENT)
+	    goto force;
+
+	/* No pid file means that we have to search in /proc/ */
     }
 
     if (pid_file && !st.st_size) {
 	warn("Empty pid file %s for %s\n", pid_file, fullname);
 
-	/* No pid file means that we have to search in /proc/ */
 	free(pid_file);
 	pid_file = NULL;
+
+	if (force)
+	    goto force;
+
+	/* No pid file means that we have to search in /proc/ */
     }
 
     if (pid_file) {		/* The case of having a pid file */
@@ -299,7 +418,7 @@ int main(int argc, char **argv)
     /* Check and verify the ignore file */
     for (c = 0; (c < iargc) && (ignore_file = iargv[c]); c++) {
 	errno = 0;
-	if (stat(ignore_file, &st) < 0) {
+	if (statn(ignore_file, STATX_SIZE, &st) < 0) {
 	    if (errno != ENOENT)
 		warn("Can not stat %s: %s\n", ignore_file, strerror(errno));
 	    continue;
@@ -317,13 +436,15 @@ int main(int argc, char **argv)
     	clear_pids();		/* Remove all pids which should be ignored */
 
     /* Do main work */
-    if (!remember) {		/* No process found with pid file */
+    if (list_empty(&remember)) {	/* No process found with pid file */
+	if (force)
+	    goto force;
         if (pidof(fullname,root,flags) < 0)
 	    exit(LSB_PROOFX);
     	clear_pids();		/* Remove all pids which should be ignored */
     }
 
-    if (remember)
+    if (!list_empty(&remember))
 	exit(LSB_OK);		/* Accordingly to LSB we have succeed. */
 
 force:
@@ -339,7 +460,7 @@ check_again:
 	exit(LSB_PROOFX);
     clear_pids();		/* Remove all pids which should be ignored */
 
-    if (!remember)
+    if (list_empty(&remember))
 	exit(LSB_NOPROC);
 
     if (seconds > 0) {
@@ -348,8 +469,16 @@ check_again:
 	goto check_again;
     }
 
-    if (!quiet)
-	    printf("%d\n", remember->pid);
+    if (!quiet) {
+	    list_t *list;
+	    int nl = 0;
+	    list_for_each(list, &remember) {
+		PROC *proc = list_entry(list, PROC);
+		if (nl++) putchar(' ');
+		printf("%ld", (long int)proc->pid);
+	    }
+	    if (nl) putchar('\n');
+    }
 
     exit(LSB_OK);
 
@@ -367,6 +496,8 @@ static int do_start(const char *inname, char *argv[], const char* log_file,
     const char * fullname;
     char proc_exe[6+9+4+1];
     static struct stat itsme;
+    sigset_t newset, oldset;
+    int fdpipe[2];
 
     if ((n = snprintf(proc_exe, sizeof(proc_exe) - 1, "/proc/%d/exe", getpid())) > 0) {
 	proc_exe[n] = '\0';
@@ -383,7 +514,7 @@ static int do_start(const char *inname, char *argv[], const char* log_file,
 
     if (log_file) {
 	errno = 0;
-	if ((tmp = fopen(log_file,"a")) == NULL)
+	if ((tmp = fopen(log_file, "a")) == NULL)
 	    error(LSB_PROOF," cannot open %s: %s\n", log_file, strerror(errno));
     }
 
@@ -391,30 +522,56 @@ static int do_start(const char *inname, char *argv[], const char* log_file,
     fflush(stderr);		/* flush stdout and especially stderr */
     errno = 0;
 
+    /*
+     * When used to start service in the init script, update the init
+     * script pid to ours first ...
+     */
+    fwd_sd_listen_pid();
+
     if (sdaemon)
 	pid = 0;
     else {
+	sigemptyset(&newset);
+	sigaddset(&newset, SIGQUIT);
+	sigaddset(&newset, SIGCHLD);
+	sigprocmask(SIG_UNBLOCK, &newset, &oldset);
 	save_sigquit = signal(SIGQUIT, sig_quit);
 	if (sigchld)
 	    (void)signal(SIGCHLD, sig_chld);
 	else
 	    (void)signal(SIGCHLD, SIG_DFL);
-    	pid = fork();
+	if (pipe2(fdpipe, O_CLOEXEC) < 0) {
+	    if (errno == ENOSYS) {
+		if (pipe(fdpipe) < 0 ||
+		    fcntl(fdpipe[0], F_SETFD, FD_CLOEXEC) < 0 ||
+		    fcntl(fdpipe[1], F_SETFD, FD_CLOEXEC) < 0)
+		    error(100, "cannot open a pipe: %m\n");
+	    }
+	}
+	if ((pid = fork()) == 0) {
+	    /* Update again to point to the child pid */
+	    fwd_sd_listen_pid();
+	}
     }
 
     switch (pid) {
     case 0:
     	if (!sdaemon) {
+	    sigprocmask(SIG_SETMASK, &oldset, NULL);
 	    (void)signal(SIGINT,  SIG_DFL);
 	    (void)signal(SIGQUIT, SIG_DFL);
 	    (void)signal(SIGSEGV, SIG_DFL);
 	    (void)signal(SIGTERM, SIG_DFL);
+
+	    close(fdpipe[1]);
+	    read(fdpipe[0], proc_exe, 1);		/* Wait on parent with the pipe here */
+	    close(fdpipe[0]);
 	}
 
 	if (root) {
 	    if (chroot(root) < 0) {
 		int lsb = LSB_WRGSYN;
-		if (errno == EPERM)
+		if (errno == EPERM || errno == EACCES)
 		    lsb = LSB_NOPERM;
 		error(lsb,"Can not change root directory to %s: %s\n", optarg, strerror(errno));
 	    }
@@ -425,6 +582,8 @@ static int do_start(const char *inname, char *argv[], const char* log_file,
 	 else
 	    set_environ(fullname);
 
+	if (schedclass >= 0)
+	    ioprio_setpid(0, schedclass, scheddata);
 	if (nicelvl != 0) {
 	    errno = 0;
 	    if (setpriority(PRIO_PROCESS, getpid(), nicelvl) < 0)
@@ -466,7 +625,7 @@ static int do_start(const char *inname, char *argv[], const char* log_file,
 	    char * redirect;
 	    if (!(redirect = getenv("REDIRECT")))
 		redirect = "/dev/tty";
-	    if ((tty = open(redirect,O_WRONLY|O_NONBLOCK,0)) < 0)
+	    if ((tty = open(redirect,O_RDWR|O_NONBLOCK,0)) < 0)
 		error(LSB_PROOF," cannot open %s: %s\n", redirect, strerror(errno));
 	    dup2(tty, fileno(stdin));
 	    dup2(tty, fileno(stdout));
@@ -500,7 +659,8 @@ static int do_start(const char *inname, char *argv[], const char* log_file,
 	fflush(stdout);
 	fflush(stderr);		/* flush stdout and especially stderr */
 	closelog();
-	chdir("/");
+	if (chdir("/") < 0)
+	    warn("error in chdir: %s\n", strerror(errno));
 	errno = 0;
 #if DEBUG
 	printf("execve(%s, [", fullname);
@@ -560,8 +720,8 @@ static int do_start(const char *inname, char *argv[], const char* log_file,
 	    fclose(tmp);
 	fflush(stdout);
 	fflush(stderr);		/* flush stdout and especially stderr */
-	usleep(1);		/* Force the kernel to run the scheduler and update
-				   the environment of the current processes */
+
+	close(fdpipe[0]);
 
 	if ((n = snprintf(proc_exe, sizeof(proc_exe) - 1, "/proc/%d/exe", pid)) > 0) {
 	    proc_exe[n] = '\0';
@@ -573,17 +733,21 @@ static int do_start(const char *inname, char *argv[], const char* log_file,
 	     * pid but before the execve() is done by the kernel, in later
 	     * case be sure not to run on our own binary.
 	     */
+	    n = 0;
 	    do {
 		struct stat serv;
 
 		errno = 0;
-		if (stat(proc_exe, &serv) < 0) {
+		if (statn(proc_exe, STATX_INO, &serv) < 0) {
 		    if (errno == ENOENT)
-			break;			/* Seems to be a fast system */
-
+			break;			/* Seems to be a very fast system
+						 * should not happen due to the pipe */
 		    error(100, "cannot stat %s: %s\n", proc_exe, strerror(errno));
 		}
 
+		if (n++ == 0)
+		    close(fdpipe[1]);		/* Sync child over the pipe */
+
 		if (itsme.st_dev != serv.st_dev || itsme.st_ino != serv.st_ino)
 		    break;			/* Seems to be a slow system */
 
@@ -592,17 +756,20 @@ static int do_start(const char *inname, char *argv[], const char* log_file,
 	    } while (true);
 
 	} else {
+
+	    close(fdpipe[1]);			/* Sync child over the pipe */
+
 	    warn("error in snprintf: %s\n", strerror(errno));
 	    usleep(100*1000);
 	}
 	n = 0;
 retry:
 	errno = 0;
-	switch (waitpid(pid, &status, WNOHANG|WUNTRACED)) {
+	switch (waitpid(pid, &status, wpopts)) {
 	case -1:		/* WNOHANG and hopefully no child but daemon */
 	    if (errno == EINTR)
 		goto retry;
-	    if (errno != ECHILD) /* ECHILD shouldnt happen, should it? (it does) */
+	    if (errno != ECHILD) /* ECHILD should not happen, should it? (it does) */
 		error(LSB_PROOFE," waitpid on %s: %s\n", fullname, strerror(errno));
 	    break;
 	case 0:			/* WNOHANG and no status available */
@@ -612,7 +779,7 @@ retry:
 	     * to see a process damage.
 	     */
 	    usleep(10*1000);	/* 10 ms time for the child and its child */
-	    if (++n < 10)
+	    if (++n < 50)
 		goto retry;
 	    break;
 	default:
@@ -630,6 +797,8 @@ retry:
 	    }
 	    break;
 	}
+	if ((wpopts & WNOHANG) && wlist)
+	    waiton(wlist);
 	break;
     }
     return 0;
@@ -640,22 +809,43 @@ retry:
  */
 static void closefds(FILE *not)
 {
+    const int fdnot = not ? fileno(not) : -1;
+    const int fderr = fileno(stderr);
+    const int sdfds = get_sd_listen_fds();
     char dir[128];
     struct dirent *fdd;
-    int fd, ret;
     DIR *fds;
+    int ret;
 
-    if (((ret = snprintf(dir, sizeof(dir), "/proc/%d/fd", (int)getpid())) < 0) ||
+    if (((ret = snprintf(dir, sizeof(dir), "/proc/%ld/fd", (long)getpid())) < 0) ||
 	(ret == sizeof(dir)))
 	error(100, "error in snprintf: %s\n", strerror(errno));
     if ((fds = opendir(dir)) == (DIR*)0)
 	error(100, "cannot open dir: %s\n", strerror(errno));
 
     while ((fdd = readdir(fds))) {
-	if ((fd = atoi(fdd->d_name)) <= fileno(stderr))
+	int fd;
+	if (*fdd->d_name == '.')
 	    continue;
-	if (not && (fd == fileno(not)))
+	if ((fd = atoi(fdd->d_name)) <= fderr) {
+	    if ((ret = fcntl(fd, F_GETFD)) < 0)
+		continue;
+	    fcntl(fd, F_SETFD, ret & ~FD_CLOEXEC);
 	    continue;
+	}
+	if (fd == fdnot) {
+	    if ((ret = fcntl(fd, F_GETFD)) < 0)
+		continue;
+	    fcntl(fd, F_SETFD, ret & ~FD_CLOEXEC);
+	    continue;
+	}
+	if (sdfds > 0 && fd >= SD_LISTEN_FDS_START
+		      && fd <  SD_LISTEN_FDS_START + sdfds) {
+	    if ((ret = fcntl(fd, F_GETFD)) < 0)
+		continue;
+	    fcntl(fd, F_SETFD, ret & ~FD_CLOEXEC);
+	    continue;
+	}
 	if (isatty(fd)) {
 	    close(fd);
 	    continue;
@@ -668,3 +858,179 @@ static void closefds(FILE *not)
     closedir(fds);
 }
 
+typedef struct _wait_
+{
+    struct _wait_ * prev;
+    struct _wait_ * next;
+    char *restrict name;
+    int wd;
+} wait_t;
+
+static void waiton(const char *list)
+{
+    int fd = inotify_init1(IN_CLOEXEC);
+    char *buf = strdup(list);
+    char *bufp;
+    wait_t *restrict p = (wait_t*)0, *n, *l, *wait = (wait_t*)0;
+
+    if (fd < 0)
+	error(100, "error in inotify_init(): %s\n", strerror(errno));
+    if (!buf)
+	error(100, "error in strdup(): %s\n", strerror(errno));
+
+    for (bufp = strsep(&buf, ":"); bufp && *bufp; bufp = strsep(&buf, ":")) {
+	char * base, * name;
+	struct stat st;
+	size_t nlen;
+
+	if (stat(bufp, &st) == 0)
+	    continue;
+	base = basename(bufp);
+	name = dirname(bufp);
+
+	if (stat(name, &st) < 0) {
+	    warn("%s: %s\n", bufp, strerror(errno));
+	    continue;
+	}
+	nlen = strlen(base);
+
+	if (posix_memalign((void*)&p, sizeof(void*), alignof(wait_t)+(nlen+1)) != 0)
+	    error(100, "malloc(): %s\n", strerror(errno));
+	p->name = ((char*)p)+alignof(wait_t);
+	strcpy(p->name, base);
+
+	if ((p->wd = inotify_add_watch(fd, name, IN_CREATE|IN_MOVE|IN_DELETE_SELF)) < 0) {
+	    warn("cannot add watch point for %s: %s\n", name, strerror(errno));
+	    free(p);
+	    continue;
+	}
+
+	if (wait)
+	    wait->prev = p;
+	p->next = wait;
+	p->prev = (wait_t*)0;
+	wait = p;
+    }
+    free(buf);
+
+    while (wait) {
+	char buf[sizeof(struct inotify_event)+PATH_MAX+1];
+	struct inotify_event *restrict ie;
+	struct timeval tv = {0, 10000};
+	fd_set check;
+	ssize_t ret;
+	int in;
+
+	FD_ZERO (&check);
+	FD_SET (fd, &check);
+	if ((in = select(fd + 1, &check, (fd_set*)0, (fd_set*)0, &tv)) < 0) {
+	    warn("select: %s\n", strerror(errno));
+	    break;
+	}
+	if (in == 0)
+	    continue;
+
+	ioctl(fd, FIONREAD, &in);
+	if (in == 0)
+	    continue;
+
+	ret = read(fd, &buf, in);
+	if (ret < 0 && (errno == EINTR))
+	    continue;
+	ie = (struct inotify_event*)&buf;
+
+	n = wait;
+	l = (wait_t*)0;
+	for (p = wait; n; p = n) {
+	    l = p->prev;
+	    n = p->next;
+
+	    if (p->wd != ie->wd)
+		continue;
+	    if (strcmp(p->name, ie->name) != 0)
+		continue;
+
+	    if (p == wait) {
+		if (n) n->prev = (wait_t*)0;
+		wait = n;
+	    } else if (l) {
+		if (n) n->prev = l;
+		l->next = n;
+	    }
+	    free(p);
+	}
+    }
+
+    close(fd);
+}
+
+static int get_sd_listen_fds()
+{
+    const char *env;
+    char *ptr = (char*)0;
+    long l;
+
+    if ((env = getenv("LISTEN_PID")) == (const char*)0)
+	return 0;
+
+    errno = 0;
+    l = strtol(env, &ptr, 10);
+    if (errno != 0)
+	return -errno;
+    if (ptr == env)
+	return -EINVAL;
+    if (*ptr != '\0')
+	return -EINVAL;
+    if (l < 0)
+	return -EINVAL;
+
+    if (getpid() != (pid_t)l)
+	return 0;
+
+    if ((env = getenv("LISTEN_FDS")) == (const char*)0)
+	return 0;
+
+    errno = 0;
+    l = strtol(env, &ptr, 10);
+    if (errno != 0)
+	return -errno;
+    if (ptr == env)
+	return -EINVAL;
+    if (*ptr != '\0')
+	return -EINVAL;
+    if (l < 0)
+	return -EINVAL;
+
+    return (int)l;
+}
+
+static void fwd_sd_listen_pid(void)
+{
+    const char *env;
+
+    /*
+     * fork & systemd socket activation:
+     * fetch listen pid and update to ours,
+     * when it is set to pid of our parent.
+     */
+    if ((env = getenv("LISTEN_PID"))) {
+	char *ptr;
+	long l;
+
+	errno = 0;
+	l = strtol(env, &ptr, 10);
+	if (errno != 0)
+	    return;
+	if (ptr == env)
+	    return;
+	if (*ptr != '\0')
+	    return;
+	if (l < 0)
+	    return;
+	if (getppid() == (pid_t)l) {
+	    char buf[24];
+	    snprintf(buf, sizeof(buf), "%d", getpid());
+	    setenv("LISTEN_PID", buf, 1);
+	}
+    }
+}
